package com.uncc.tv.recommender;

import com.lhy.tool.ToolException;
import com.lhy.tool.util.Utils;
import com.uncc.tv.ParamManager;
import com.uncc.tv.TVConstants;
import com.uncc.tv.TVException;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.apache.commons.math3.linear.BlockRealMatrix;
import org.apache.commons.math3.linear.LUDecomposition;

/**
 * This is Mixture Probabilistic Matrix Factorization.
 * @author huayuli
 *
 */

public class GPMF extends PMFBase {
	protected static final String FILE_NAME_VAR_PHI        = "Var_Phi";
	protected static final String FILE_NAME_VAR_WEIGHT     = "Var_Weight";
	protected static final String FILE_NAME_VAR_SIGMA      = "Var_Sigma";
	protected static final String FILE_NAME_VAR_MU         = "Var_Mu";
	protected static final String FILE_NAME_SIGMA_DET_SQRT = "SigmaDetSqrt";

	protected float[] varWeight             = null;
	protected float[] varPhi                = null;
	protected float[] varMu                 = null;
	protected float[] sigmaDetSqrt          = null;
	protected boolean[] weightAssignments   = null;
	// the first element is 0, the element is the current group television start index
	protected int[] televisionGroupNumIndex = null;
	protected int[] televisionNumInGroup    = null;
	protected int[] groupMixtureNums        = null;
	protected int GROUP_NUM;
	protected int TOTAL_LATENT_USER_NUM;
	protected int TOTAL_MIXTURE_NUM;

	protected float lambdaRegT      = 0.001f;
	protected float lambda          = 1f;
	private final float lambdaSigma = 15;

	protected float alpha        = 1.02f;
	protected float beta0        = 0.5f;
	protected float v0_D_1       = -1f;
	protected float[] W0_Inverse = null;

	protected File sigmaFile          = null;
	protected boolean isUseWeightInfo = true;

	public GPMF(ParamManager paramManager)
			throws TVException, IOException {
		this(paramManager, "GPMF", false);
	}

	public GPMF(ParamManager paramManager, String programName,
			boolean forceNoMinusMeanRating) throws TVException,
							IOException {
		PROGRAM_NAME = programName;
		if (forceNoMinusMeanRating) {
			MINUS_MEAN_RATING = false;
			Utils.println("[Info] Force not minus mean rating.");
		}

		init(paramManager);
	}

	public void estimate() throws TVException, IOException {
		BufferedWriter resultWriter = Utils.createBufferedWriter(resultFile, true);
		getParams().store(resultWriter, String.format("%s Parameters", PROGRAM_NAME));

		double oldlowerbound  = calculateLowerbound();
		double errorThreshold = Double.MAX_VALUE;
		int iteration         = 1;
		if (testFile != null) {
			Utils.writeAndPrint(resultWriter, String.format(
				"Train RMSE = %s, Test RMSE = %s",
				predict(trainFile), predict(testFile)), true);
		} else {
			Utils.writeAndPrint(resultWriter,
				"Train RMSE = " + predict(trainFile), true);
		}
		do {
			updateProgramFeature();
			updateTelevisionFeature();
			updateWeight();
			updatePhi();
			updateMu();
			//if (iteration == 1) {
				//updateSigma();
			//}
			double lowerbound = calculateLowerbound();
			errorThreshold    = Math.abs(Math.abs((oldlowerbound - lowerbound) / oldlowerbound));
			oldlowerbound     = lowerbound;
			if (testFile != null) {
				Utils.writeAndPrint(resultWriter, String.format(
					"[Iteration = %s] LowerBound = %s, ErrorThreshold = %s, TrainRMSE = %s, TestRMSE = %s",
					iteration, lowerbound, errorThreshold,
					predict(trainFile), predict(testFile)),
					true);
			} else {
				Utils.writeAndPrint(resultWriter, String.format(
					"[Iteration = %s] LowerBound = %s, ErrorThreshold = %s, TrainRMSE = %s",
					iteration, lowerbound, errorThreshold,
					predict(trainFile)),
					true);
			}
			iteration ++;
		} while (iteration < 5 || errorThreshold > ERROR_THRESHOLD);
		storeLatentVariables(outputPath, true);
		resultWriter.close();
	}

	protected void updateTelevisionFeature() throws IOException, TVException {
		BufferedReader sigmaReader = Utils.createBufferedReader(sigmaFile);
		int televisionIndex = 0;
		int televisionId    = 0;
		int recordIndex     = 0;
		int phiIndex        = 0;
		int muStartIndex    = 0;
		float[][] sigma = null;
		for (int groupId = 0; groupId < GROUP_NUM; groupId ++) {
			int K = groupMixtureNums[groupId];
			sigma = new float[K][FEATURE_NUM * FEATURE_NUM];
			for (int k = 0; k < K; k ++) {
				sigma[k] = readDataInLine(sigmaReader);
			}
			for (int innerTelevisionId = 0; innerTelevisionId <
					televisionNumInGroup[groupId]; innerTelevisionId ++) {
				double[][] VVT = new double[FEATURE_NUM][FEATURE_NUM];
				float[]      T = new float[FEATURE_NUM];
				for (int i = 0; i < programNumInTelevisionOrder[televisionId]; i ++) {
					int programStartIndex =
						programIndexInTelevisionOrder[recordIndex] *
						FEATURE_NUM;
					for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
						float currentValue = programFeature[programStartIndex + featureId];
						T[featureId] += ratingsInTelevisionOrder[recordIndex] *
								currentValue;
						for (int f = 0; f < FEATURE_NUM; f ++) {
							VVT[featureId][f] += currentValue *
								programFeature[programStartIndex + f];
						}
					}
					recordIndex ++;
				}
				for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
					T[featureId] *= lambda;
					for (int f = 0; f < FEATURE_NUM; f ++) {
						VVT[featureId][f] *= lambda;
					}
					VVT[featureId][featureId] += lambdaRegT;
				}
				for (int k = 0; k < K; k ++) {
					int tmpMuStartIndex = muStartIndex + k * FEATURE_NUM;
					for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
						float sum           = 0;
						int sigmaStartIndex = featureId * FEATURE_NUM;
						for (int f = 0; f < FEATURE_NUM; f ++) {
							float value = varPhi[phiIndex] * sigma[k][sigmaStartIndex + f];
							VVT[featureId][f] += value;
							sum += value * varMu[tmpMuStartIndex + f];
						}
						T[featureId] += sum;
					}
					phiIndex ++;
				}
				VVT = inverse(VVT);
				for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
					televisionFeature[televisionIndex] = 0;
					for (int f = 0; f < FEATURE_NUM; f ++) {
						televisionFeature[televisionIndex] +=
							VVT[featureId][f] * T[f];
					}
					televisionIndex ++;
				}
				televisionId ++;
			} // end for (int innerTelevisionId = 0
			muStartIndex += K * FEATURE_NUM;
		} // end for (int groupId = 0;
		String line = null;
		if ((line = sigmaReader.readLine()) != null)
			throw new TVException("Sigma file does not end. NextLine:%s", line);
		sigmaReader.close();
	}

	protected void updateWeight() throws IOException, TVException {
		int muStartIndex = 0;
		for (int groupId = 0; groupId < GROUP_NUM; groupId ++) {
			int K = groupMixtureNums[groupId];
			for (int innerTelevisionId = 0; innerTelevisionId <
					televisionNumInGroup[groupId]; innerTelevisionId ++) {
				float sumPhi = 0;
				for (int k = 0; k < K; k ++) {
					if (weightAssignments == null ||
							weightAssignments[muStartIndex + k]) {
						sumPhi += (
							varWeight[muStartIndex + k] = varPhi[muStartIndex + k] + alpha - 1
							);
					}
				}
				if (sumPhi <= 0) {
					throw new TVException("UpdateWeigth: sumPhi = %s",
							sumPhi);
				}
				for (int k = 0; k < K; k ++) {
					varWeight[muStartIndex + k] /= sumPhi;
				}
				muStartIndex += K;
			}
		}
	}
	
	protected void updatePhi() throws IOException, TVException {
		BufferedReader sigmaReader = Utils.createBufferedReader(sigmaFile);
		int sigmaDetStartIndex    = 0;
		int phiStartIndex         = 0;
		int muStartIndex          = 0;
		int televisionStartIndex  = 0;
		for (int groupId = 0; groupId < GROUP_NUM; groupId ++) {
			int K = groupMixtureNums[groupId];
			float[][] sigmas = new float[K][FEATURE_NUM * FEATURE_NUM];
			for (int k = 0; k < K; k ++) {
				sigmas[k] = readDataInLine(sigmaReader);
			}
			for (int innerTelevisionId = 0; innerTelevisionId <
					televisionNumInGroup[groupId]; innerTelevisionId ++) {
				float phiSum = 0;
				for (int k = 0; k < K; k ++) {
					float sigmaDS = sigmaDetSqrt[sigmaDetStartIndex + k];
					float[] diff  = new float[FEATURE_NUM];
					for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
						diff[featureId] =
							televisionFeature[televisionStartIndex + featureId] -
							varMu[muStartIndex + k * FEATURE_NUM + featureId];
					}
					float exp = 0;
					for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
						float v = 0;
						for (int f = 0; f < FEATURE_NUM; f ++) {
							v += diff[f] * sigmas[k][f * FEATURE_NUM + featureId];
						}
						exp += v * diff[featureId];
					}
					if (varWeight[phiStartIndex + k] == 0) {
						varPhi[phiStartIndex + k] = 0;
					} else {
						varPhi[phiStartIndex + k] = (float)(Math.exp(sigmaDS) *
							varWeight[phiStartIndex + k] * Math.exp(-0.5 * exp));
						phiSum += varPhi[phiStartIndex + k];
					}
				}
				if (phiSum <= 0) {
					throw new TVException("UpdatePhi: phiSum = %s", phiSum);
				}
				if (Double.isInfinite(phiSum) || Double.isNaN(phiSum)) {
					normalize(varPhi, phiStartIndex, K);
				} else {
					for (int k = 0; k < K; k ++) {
						varPhi[phiStartIndex + k] /= phiSum;
					}
				}
				phiStartIndex   += K;
				televisionStartIndex += FEATURE_NUM;
			}
			sigmaDetStartIndex += K;
			muStartIndex       += K * FEATURE_NUM;
		}
		String line = null;
		if ((line = sigmaReader.readLine()) != null)
			throw new TVException("Sigma file does not end. NextLine:%s", line);
		sigmaReader.close();
	}

	

	protected void updateMu() throws IOException, TVException {
		int phiStartIndex = 0;
		int muIndex       = 0;
		for (int groupId = 0; groupId < GROUP_NUM; groupId ++) {
			int K = groupMixtureNums[groupId];
			for (int k = 0; k < K; k ++) {
				float[] avgT = new float[FEATURE_NUM];
				float phiSum = 0.0f;
				for (int innerTelevisionId = 0; innerTelevisionId <
						televisionNumInGroup[groupId]; innerTelevisionId ++) {
					int televisionStartIndex = (televisionGroupNumIndex[groupId] +
							innerTelevisionId) * FEATURE_NUM;
					float phiValue = varPhi[phiStartIndex + innerTelevisionId * K + k];
					phiSum += phiValue;
					for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
						avgT[featureId] += phiValue * televisionFeature[televisionStartIndex + featureId];
					}
				}
				for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
					varMu[muIndex] = avgT[featureId] / (beta0 + phiSum);
					muIndex ++;
				}
			}
			phiStartIndex += televisionNumInGroup[groupId] * K;
		}
	}

	// Sigma only contains one element
	private void updateSigma() throws IOException, TVException {
		BufferedWriter sigmaWriter = Utils.createBufferedWriter(sigmaFile);
		int phiStartIndex          = 0;
		int muStartIndex           = 0;
		int sigmaDetSqrtIndex      = 0;
		for (int groupId = 0; groupId < GROUP_NUM; groupId ++) {
			int K    = groupMixtureNums[groupId];
			int TNum = televisionNumInGroup[groupId];
			if (TNum <= 1) {
				for (int k = 0; k < K; k ++) {
					for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
						for (int f = 0; f < FEATURE_NUM; f ++) {
							if (featureId == f) {
								sigmaWriter.write("1");
							} else {
								sigmaWriter.write("0");
							}
							if (f != FEATURE_NUM - 1 ||
									featureId != FEATURE_NUM - 1) {
								sigmaWriter.write(" ");
							}
						}
					}
					if (groupId != TOTAL_MIXTURE_NUM - 1 || k != K -1) {
						sigmaWriter.newLine();
					}
					sigmaDetSqrt[sigmaDetSqrtIndex] = 0;
					sigmaDetSqrtIndex ++;
					muStartIndex += FEATURE_NUM;
				}
			} else {
			for (int k = 0; k < K; k ++) {
				double[][] sigma = new double[FEATURE_NUM][FEATURE_NUM];
				float phiSum     = 0;
				float slk        = 0;
				for (int innerTelevisionId = 0; innerTelevisionId <TNum; innerTelevisionId ++) {
					int tStartIndex = (televisionGroupNumIndex[groupId] +
						innerTelevisionId) * FEATURE_NUM;
					float phiValue = varPhi[phiStartIndex + innerTelevisionId * K + k];
					float diff[]   = new float[FEATURE_NUM];
					for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
						diff[featureId] =
							televisionFeature[tStartIndex + featureId] -
							varMu[muStartIndex + featureId];
					}
					for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
						slk += diff[featureId] * diff[featureId] * phiValue;
					}
					phiSum += phiValue;
				}
				for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
					slk += beta0 * varMu[muStartIndex + featureId] *
						varMu[muStartIndex + featureId];
				}
				for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
					slk += W0_Inverse[featureId] + lambdaSigma;
				}
				slk = FEATURE_NUM *(v0_D_1 + 1 + phiSum) / slk;
				if (slk <= 1.0e-7f) slk = 1.0e-7f;
				sigmaDetSqrt[sigmaDetSqrtIndex] = (float)(0.5*FEATURE_NUM*Math.log(slk));
				for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
					for (int f = 0; f < FEATURE_NUM; f ++) {
						if (featureId == f) {
							sigmaWriter.write(String.valueOf(slk));
						} else {
							sigmaWriter.write("0");
						}
						if (f != FEATURE_NUM - 1 ||
								featureId != FEATURE_NUM - 1) {
							sigmaWriter.write(" ");
						}
					}
				}
				if (groupId != TOTAL_MIXTURE_NUM - 1 || k != K -1) {
					sigmaWriter.newLine();
				}
				if (Double.isNaN(sigmaDetSqrt[sigmaDetSqrtIndex]) ||
						Double.isInfinite(sigmaDetSqrt[sigmaDetSqrtIndex])) {
					Utils.err("sigmaDetSqrt[sigmaDetSqrtIndex]=" + sigmaDetSqrt[sigmaDetSqrtIndex]);
					for (int innerTelevisionId = 0; innerTelevisionId <TNum; innerTelevisionId ++) {
						float phiValue = varPhi[phiStartIndex + innerTelevisionId * K + k];
						System.out.print(phiValue + "\t");
					}
					System.out.println();
					Utils.save(sigma, TVConstants.SPACER,
						new File(outputPath, "sigma"), true);
					try {
					Utils.println("sqrt:" + Math.sqrt(new LUDecomposition(
							new BlockRealMatrix(Utils.load2Array(
							new File(outputPath, "nrmSigma"), "\t"))).getDeterminant()));
					} catch (ToolException e) {
						e.printStackTrace();
					}
					System.exit(-1);
					
				}
				sigmaDetSqrtIndex ++;
				muStartIndex += FEATURE_NUM;
			} // end for k
			}
			phiStartIndex += TNum * K;
		} // end for groupId
		sigmaWriter.close();
	}

	protected double calculateLowerbound() throws IOException, TVException {
		double lbvalue   = 0;
		// (R - T'V)_2
		int recordIndex = 0;
		for (int televisionId = 0; televisionId < TELEVISION_NUM; televisionId ++) {
			int televisionStartIndex = televisionId * FEATURE_NUM;
			for (int i = 0; i < programNumInTelevisionOrder[televisionId]; i ++) {
				double v = 0;
				int programStartIndex =
					programIndexInTelevisionOrder[recordIndex] *
					FEATURE_NUM;
				for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {					
					v += televisionFeature[televisionStartIndex + featureId] *
						programFeature[programStartIndex + featureId];
				}
				lbvalue += (ratingsInTelevisionOrder[recordIndex] - v) *
					(ratingsInTelevisionOrder[recordIndex] - v);
				recordIndex ++;
			}
		}
		lbvalue *= -0.5 * lambda;

		// V'V
		double sum       = 0;
		int programIndex = 0;
		for (int programId = 0; programId < PROGRAM_NUM; programId ++) {
			for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
				sum += programFeature[programIndex] *
					programFeature[programIndex];
				programIndex ++;
			}
		}
		lbvalue += -0.5 * lambdaProgram * sum;
		if (Double.isNaN(lbvalue) || Double.isInfinite(lbvalue)) {
			Utils.err("V'V : lbvalue = " + lbvalue);
		}

		// T
		BufferedReader sigmaReader = Utils.createBufferedReader(sigmaFile);
		double phiSum               = 0;
		double phiLnPhiSum          = 0;
		double phiLnWeightSum       = 0;
		double phiLnSigmaDetSqrtSum = 0;
		double lnSigmaDetSqrtSum    = 0;
		double lnWeightSum          = 0;
		double muTmu                = 0;
		double trTW0Sigma           = 0;
		double T_MuTT_Mu            = 0;
		int phiStartIndex          = 0;
		int lkIndex                = 0;
		int muStartIndex           = 0;
		int televisionStartIndex   = 0;
		for (int groupId = 0; groupId < GROUP_NUM; groupId ++) {
			int K = groupMixtureNums[groupId];
			for (int k = 0; k < K; k ++) {
				float[] sigma = readDataInLine(sigmaReader);
				for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
					//muTmu      += varMu[muStartIndex + featureId] * varMu[muStartIndex + featureId];
					for (int f = 0; f < FEATURE_NUM; f ++) {
						muTmu += varMu[muStartIndex + f] *
							sigma[f * FEATURE_NUM + featureId] *
							varMu[muStartIndex + featureId];
					}
					trTW0Sigma += W0_Inverse[featureId] *
						sigma[featureId * FEATURE_NUM + featureId];
				}
				double phiSumTelevision = 0;
				for (int innerTelevisionId = 0; innerTelevisionId <
						televisionNumInGroup[groupId]; innerTelevisionId ++) {
					int phiIndex = phiStartIndex + innerTelevisionId * K + k;
					double phi        = varPhi[phiIndex];
					double weight     = varWeight[phiIndex];
					if (weightAssignments == null ||
							weightAssignments[phiIndex]) {
						phiLnWeightSum   += phi * Math.log(weight);
						lnWeightSum      += Math.log(weight);
						phiSumTelevision += phi;
						phiLnPhiSum      += phi * Math.log(phi);
					}
					//
					float[] diff = new float[FEATURE_NUM];
					int tIndex   = televisionStartIndex + innerTelevisionId * FEATURE_NUM;
					for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
						diff[featureId] = televisionFeature[tIndex + featureId] -
								varMu[muStartIndex + featureId];
					}
					double TTsum = 0;
					for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
						double v = 0;
						for (int f = 0; f < FEATURE_NUM; f ++) {
							v += diff[f] * sigma[f * FEATURE_NUM + featureId];
						}
						TTsum += v * diff[featureId] ;
					}
					T_MuTT_Mu += TTsum * phi;
				}
				lnSigmaDetSqrtSum    += sigmaDetSqrt[lkIndex];
				phiSum               += phiSumTelevision;
				phiLnSigmaDetSqrtSum += sigmaDetSqrt[lkIndex] *
						phiSumTelevision;
				muStartIndex         += FEATURE_NUM;
				lkIndex ++;
			} // end for k
			phiStartIndex        += televisionNumInGroup[groupId] * K;
			televisionStartIndex += televisionNumInGroup[groupId] * FEATURE_NUM;
		}
		lbvalue += -0.5 * FEATURE_NUM * Math.log(2 * Math.PI) * phiSum;
		if (Double.isNaN(lbvalue) || Double.isInfinite(lbvalue)) {
			Utils.err("-0.5 * FEATURE_NUM * Math.log(2 * Math.PI) * phiSum : lbvalue = " + lbvalue);
		}
		lbvalue += (alpha - 1) * lnWeightSum;
		if (Double.isNaN(lbvalue) || Double.isInfinite(lbvalue)) {
			Utils.err("(alpha - 1) * Math.log(lnWeightSum) : lbvalue=%s, lnWeightSum=%s", lbvalue, lnWeightSum);
		}
		lbvalue += phiLnWeightSum;
		lbvalue -= phiLnPhiSum;
		lbvalue += (v0_D_1 + 1) * lnSigmaDetSqrtSum;
		if (Double.isNaN(lbvalue) || Double.isInfinite(lbvalue)) {
			Utils.err("v0_D_1 * lnSigmaDetSqrtSum : lbvalue = " + lbvalue);
		}
		lbvalue -= 0.5 * muTmu * beta0;
		if (lbvalue > 0) Utils.err("v0_D_1 * lnSigmaDetSqrtSum = " + lbvalue);
		lbvalue += phiLnSigmaDetSqrtSum;
		if (Double.isNaN(lbvalue) || Double.isInfinite(lbvalue)) {
			Utils.err("0.5 * v0_D_1 * Math.log(lnSigmaDetSqrtSum) : lbvalue = " + lbvalue);
		}
		lbvalue -= 0.5 * trTW0Sigma;
		lbvalue -= 0.5 * T_MuTT_Mu;
		String line = null;
		if ((line = sigmaReader.readLine()) != null) 
			throw new TVException("Sigma file does not end. NextLine:%s", line);
		sigmaReader.close();
		return lbvalue;
	}

	public float[] readDataInLine(BufferedReader reader) throws IOException {
		String line    = reader.readLine();
		String[] array = line.split(" ");
		float[] result = new float[array.length];
		for (int i = result.length - 1; i >=0; i--) {
			result[i] = Float.parseFloat(array[i]);
		}
		return result;
	}

	protected void init(ParamManager paramManager) throws TVException, IOException {
		if (paramManager == null)
			throw new TVException("No parameters specified.");

		ERROR_THRESHOLD = paramManager.getConvergeThreshold() == null ? 1e-5f :
								paramManager.getConvergeThreshold();
		FEATURE_NUM     = paramManager.getFeatureNum();
		PROGRAM_NUM     = paramManager.getProgramNum();
		GROUP_NUM       = paramManager.getGroupNum();
		TELEVISION_NUM  = paramManager.getTelevisionNum();

		outputPath = paramManager.getOutputPath();
		if (outputPath == null)
			throw new TVException("No output path specified.");
		outputPath = new File(new File(outputPath, PROGRAM_NAME),
					String.valueOf(FEATURE_NUM));
		initPath   = new File(outputPath, "Init");
		trainFile  = paramManager.getTrainRatingFile();
		testFile   = paramManager.getTestFile();
		resultFile = new File(outputPath, "Result");
		if (paramManager.getTrainClusterFile() == null)
			throw new TVException("No train file specified.");

		sigmaFile = new File(outputPath, FILE_NAME_VAR_SIGMA);

		// it must be called before
		parseTrainClusterFile(paramManager.getTrainClusterFile());
		parseTrainRatingFile(paramManager.getTrainRatingFile());

		if (! isUseWeightInfo) weightAssignments = null;

		initProgramFeature(paramManager.getProgramFeatureInitFile()); // new File(initPath, FILE_NAME_PROGRAM_FEATURE)
		initTelevisionFeature(paramManager.getTelevisionFeatureInitFile()); // new File(initPath, FILE_NAME_PROGRAM_FEATURE)
		initWeightAndPhi();
		initMu();
		initSigma();
		initW0();

		storeParams("GPMF Parameters");
	}

	@Override
	protected void initTelevisionFeature(File televisionFeatureFile)
			throws TVException {
		if (televisionFeatureFile == null) {
			super.initTelevisionFeature(null);
		} else {
			try {
				//File televisionFeatureFile = new File(initPath, FILE_NAME_TELEVISION_FEATURE);
				Utils.println("!![Info] Loading televisionFeature from " + televisionFeatureFile.getAbsolutePath());
				/*float[] origTelevisionFeature = convert2DTo1D(Utils.load2FloatArray(
					televisionFeatureFile, TVConstants.SPACER));
				if (origTelevisionFeature.length != TELEVISION_NUM * FEATURE_NUM) {
					throw new TVException("Failed to load teleivsionFeature: dimension = %s which should be %s.",
						televisionFeature.length, TELEVISION_NUM * FEATURE_NUM);
				}*/
				float[][] origTelevisionFeature = Utils.load2FloatArray(
						televisionFeatureFile, TVConstants.SPACER);
				if (origTelevisionFeature.length != TELEVISION_NUM) {
					throw new TVException("Failed to load teleivsionFeature: dimension = %s which should be %s.",
						televisionFeature.length, TELEVISION_NUM);
				}
				televisionFeature = new float[TELEVISION_NUM * FEATURE_NUM];
				for (int realTelevisionId = 0; realTelevisionId < TELEVISION_NUM;
						realTelevisionId ++) {
					int id = realUserIdMapToTelevisionIndex[realTelevisionId] * FEATURE_NUM;
					for (int i = 0; i < FEATURE_NUM; i ++) {
						televisionFeature[id + i] = origTelevisionFeature[realTelevisionId][i];
					}
				}
			} catch (ToolException e) {
				throw new TVException(e.toString());
			}
		}
	}

	private void initSigma() throws IOException {
		sigmaDetSqrt = new float[TOTAL_MIXTURE_NUM];
		int index    = 0;
		BufferedWriter sigmaWriter = Utils.createBufferedWriter(sigmaFile);
		for (int groupId = 0; groupId < GROUP_NUM; groupId ++) {
			for (int k = 0; k < groupMixtureNums[groupId]; k ++) {
				for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
					for (int f = 0; f < FEATURE_NUM; f ++) {
						if (featureId == f) {
							sigmaWriter.write("1");
						} else {
							sigmaWriter.write("0");
						}
						if (f != FEATURE_NUM - 1 ||
							featureId != FEATURE_NUM - 1) {
							sigmaWriter.write(" ");
						}
					}
				}
				sigmaDetSqrt[index] = 0; 
				if (groupId != GROUP_NUM - 1 ||
						k != groupMixtureNums[groupId] -1) {
					sigmaWriter.newLine();
				}
				index ++;
			}
		}
		sigmaWriter.close();
	}

	/*
	 * Must call after method initTelevisionFeature()
	 */
	private void initMu() throws TVException {
		if (! INIT_FROM_FILE) {
			varMu = new float[TOTAL_MIXTURE_NUM * FEATURE_NUM];
			/*for (int i = varMu.length - 1; i >= 0; i --) {
				varMu[i] = (float)Math.random();
			}*/
			int televisionIndex = 0;
			int muIndex         = 0;
			for (int groupId = 0; groupId < GROUP_NUM; groupId ++) {
				int K        = groupMixtureNums[groupId];
				int TNums    = televisionNumInGroup[groupId];
				float mean[] = new float[FEATURE_NUM];
				for (int t = 0; t < TNums; t ++) {
					for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
						mean[featureId] += televisionFeature[televisionIndex];
						televisionIndex ++;
					}
				}
				for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
					if (TNums <= 0) throw new TVException("TNums = " + TNums);
					mean[featureId] /= TNums;
				}
				for (int k = 0; k < K; k ++) {
					for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
						varMu[muIndex] = mean[featureId];
						muIndex ++;
					}
				}
			}
		} else {
			File muFile = new File(initPath, FILE_NAME_VAR_MU);
			Utils.println("[Info] Loading VarMu from " + muFile.getAbsolutePath());
			try {
				varMu = convert2DTo1D(Utils.load2FloatArray(
						muFile, TVConstants.SPACER));
				if (varMu.length != TOTAL_MIXTURE_NUM * FEATURE_NUM) {
					throw new TVException("Failed to load varMu : dimesion = %s which should be %s.",
						varMu.length, TOTAL_MIXTURE_NUM * FEATURE_NUM);
				}
			} catch (ToolException e) {
				throw new TVException(e.toString());
			}
		}
	}

	private void initWeightAndPhi() throws TVException {
		if (! INIT_FROM_FILE) {
			varWeight = new float[TOTAL_LATENT_USER_NUM];
			varPhi    = new float[TOTAL_LATENT_USER_NUM];
			int index = 0;
			for (int groupId = 0; groupId < GROUP_NUM; groupId ++) {
				int K = groupMixtureNums[groupId];
				for (int innerTelevisionId = 0; innerTelevisionId <
						televisionNumInGroup[groupId]; innerTelevisionId ++) {
					float sumPhi = 0;
					for (int k = 0; k < K; k ++) {
						varPhi[index + k]    = (float)Math.random();
						sumPhi += varPhi[index + k];
					}
					for (int k = 0; k < K; k ++) {
						if (weightAssignments == null ||
								weightAssignments[index + k]) {
							varWeight[index + k] =
							(varPhi[index + k] + alpha - 1) /
							(sumPhi + alpha * K - K);
						} else {
							varWeight[index+k]=0;
						}
						varPhi[index + k] =
							varPhi[index + k] / sumPhi;
					}
					index += K;
				}
			}
		} else {
			try {
				File weightFile = new File(initPath, FILE_NAME_VAR_WEIGHT);
				Utils.println("[Info] Loading varWeight from " + weightFile.getAbsolutePath());
				varWeight = Utils.load1FloatArray(
					weightFile, TVConstants.SPACER);
				if (varWeight.length != TOTAL_LATENT_USER_NUM) {
					throw new TVException("Failed to load varWeigth: dimension = %s which should be %s.",
						varWeight.length, TOTAL_LATENT_USER_NUM);
				}
				File phiFile = new File(initPath, FILE_NAME_VAR_PHI);
				Utils.println("[Info] Loading varPhi from " + phiFile.getAbsolutePath());
				varPhi = Utils.load1FloatArray(
					phiFile, TVConstants.SPACER);
				if (varPhi.length != TOTAL_LATENT_USER_NUM) {
					throw new TVException("Failed to load varPhi: dimension = %s which should be %s.",
						varPhi.length, TOTAL_LATENT_USER_NUM);
				}
			} catch (ToolException e){
				throw new TVException(e.toString());
			}
		}
	}

	private void initW0() {
		// init a diagonal matrix
		W0_Inverse = new float[FEATURE_NUM];
		for (int i = FEATURE_NUM - 1; i >= 0; i --) {
			W0_Inverse[i] = 1;
		}
	}

	@Override
	protected Properties getParams() {
		Properties props = new Properties();
		props.putAll(super.getParams());
		props.put("Alpha",		String.valueOf(alpha));
		props.put("TotalLatentUserNum",	String.valueOf(TOTAL_LATENT_USER_NUM));
		props.put("GroupNum",		String.valueOf(GROUP_NUM));
		props.put("TotalMixtureNum",	String.valueOf(TOTAL_MIXTURE_NUM));
		return props;
	}

	/*
	 * The train file format is as follows:
	 * GroupId	'\t' UserId '\t' MixtureK '\t' MxitureClusterAssignments
	 * The data stores according 
	 */
	private void parseTrainClusterFile(File trainClusterFile) throws TVException {
		groupMixtureNums               = new int[GROUP_NUM];
		televisionNumInGroup           = new int[GROUP_NUM];
		realUserIdMapToTelevisionIndex = new int[TELEVISION_NUM];
		TOTAL_LATENT_USER_NUM          = 0;
		TOTAL_MIXTURE_NUM              = 0;
		List<Boolean> weightAssignList = new ArrayList<Boolean>();
		BufferedReader reader = null;
		try {
			Map<Integer, Integer> userIdArrayIndexMap =
				new HashMap<Integer, Integer>();
			reader      = Utils.createBufferedReader(trainClusterFile);
			String line = null;
			Set<Integer> groupIdSet    = new HashSet<Integer>();
			Set<Integer> realUserIdSet = new HashSet<Integer>();
			int userIndex = 0;
			int minGroupId = Integer.MAX_VALUE;
			int maxGroupId = -1;
			int minRealUserId = Integer.MAX_VALUE;
			int maxRealUserId = -1;
			while ((line = reader.readLine()) != null) {
				String[] array = line.split("\t");
				if (array.length < 3) {
					Utils.cleanup(reader);
					reader = null;
					throw new TVException("Failed to parse line: " + line);
				}
				int groupId  = Integer.parseInt(array[0]);
				int userId   = Integer.parseInt(array[1]);
				int mixtureK = Integer.parseInt(array[2]);
				if (mixtureK <= 0) {
					Utils.cleanup(reader);
					reader = null;
					throw new TVException("Failed to parse train cluster file: mixtureK = %s",
							mixtureK);
				}
				int listStartIndex = weightAssignList.size();
				for (int i = 0; i < mixtureK; i ++) {
					weightAssignList.add(false);
				}
				for (int i = 3; i < array.length; i ++) {
					int index = Integer.parseInt(array[i]);
					if (listStartIndex + index >= weightAssignList.size()) {
						reader.close();
						reader = null;
						throw new TVException("Out of range: LineNum=%s,weightLen=%s,index=%s, line:%s",
							userIndex + 1, weightAssignList.size(), listStartIndex + index, line);
					}
					weightAssignList.set(listStartIndex + index, true);
				}
				userIdArrayIndexMap.put(userId, userIndex);
				realUserIdMapToTelevisionIndex[userId] = userIndex;
				televisionNumInGroup[groupId] ++;
				TOTAL_LATENT_USER_NUM += mixtureK;
				if (! groupIdSet.contains(groupId)) {
					groupIdSet.add(groupId);
					TOTAL_MIXTURE_NUM        += mixtureK;
					groupMixtureNums[groupId] = mixtureK;
				} else
				if (groupMixtureNums[groupId] != mixtureK) {
					Utils.cleanup(reader);
					reader = null;
					throw new TVException("Failed to parse train cluster file : mixture K does not match: " + line);
				}
				if (groupId < minGroupId) minGroupId = groupId;
				if (groupId > maxGroupId) maxGroupId = groupId;
				if (! realUserIdSet.contains(userId))
					realUserIdSet.add(userId);
				if (userId < minRealUserId) minRealUserId = userId;
				if (userId > maxRealUserId) maxRealUserId = userId;
				userIndex ++;
			}
			televisionGroupNumIndex = new int[GROUP_NUM];
			for (int i = 1;i < televisionNumInGroup.length; i ++) {
				televisionGroupNumIndex[i] = televisionGroupNumIndex[i - 1] +
					televisionNumInGroup[i - 1]; 
			}
			weightAssignments = new boolean[weightAssignList.size()];
			for (int i = 0; i < weightAssignments.length; i ++) {
				weightAssignments[i] = weightAssignList.get(i);
			}
			if (groupIdSet.size() != GROUP_NUM) {
				throw new TVException("Parsed group num does not match with predefined group num.[ParsedGroupNum=%s][PredefinedGroupNum=%s]",
					groupIdSet.size(), GROUP_NUM);
			}
			if (minGroupId != 0) {
				throw new TVException("MinGroupId = %s, instead of 0.",
					minGroupId);
			}
			if (maxGroupId != groupIdSet.size() - 1) {
				throw new TVException("MaxGroupId = %s, instead of %s.",
					maxGroupId, groupIdSet.size() - 1);
			}
			if (realUserIdSet.size() != TELEVISION_NUM) {
				throw new TVException("Parsed television num does not match with predefined television num.[ParsedTelevisionNum=%s][PredefinedTelevisionNum=%s]",
					realUserIdSet.size(), TELEVISION_NUM);
			}
			if (minRealUserId != 0)   {
				throw new TVException("MinRealUserId = %s, instead of 0.",
					minRealUserId);
			}
			if (maxRealUserId != realUserIdSet.size() - 1) {
				throw new TVException("MaxRealUserId = %s, instead of %s.",
					maxRealUserId, realUserIdSet.size() - 1);
			}
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			Utils.cleanup(reader);
		}
	}
	
	// Except the sigma
	protected void storeLatentVariables(File outputPath, boolean wrapPhi) {
		// store program feature
		store2DArray(programFeature, FEATURE_NUM,
				new File(outputPath, FILE_NAME_PROGRAM_FEATURE));
		// store television feature
		store2DArray(televisionFeature, FEATURE_NUM,
				new File(outputPath, FILE_NAME_TELEVISION_FEATURE));
		// store phi
		File phiFile = new File(outputPath, FILE_NAME_VAR_PHI);
		if (wrapPhi) {
			storeTelevisionKVariables(varPhi, phiFile);
		} else {
			Utils.save(varPhi, TVConstants.SPACER, phiFile, false, false);
		}
		// store weight
		File weightFile = new File(outputPath, FILE_NAME_VAR_WEIGHT);
		if (wrapPhi) {
			storeTelevisionKVariables(varWeight, weightFile);
		} else {
			Utils.save(varWeight, TVConstants.SPACER, weightFile, false, false);
		}
		// store sigmaDetSqrt
		Utils.save(sigmaDetSqrt, TVConstants.SPACER,
				new File(outputPath, FILE_NAME_SIGMA_DET_SQRT),
				false, false);
		BufferedWriter writer = null;
		try {
			// store mu
			writer = Utils.createBufferedWriter(new File(outputPath, FILE_NAME_VAR_MU));
			int count = 0;
			for (int groupId = 0; groupId < GROUP_NUM; groupId ++) {
				int K = groupMixtureNums[groupId];
				for (int k = 0; k < K; k ++) {
					for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
						writer.write(String.valueOf(varMu[count]));
						count ++;
						if (featureId != FEATURE_NUM - 1)
							writer.write(TVConstants.SPACER);
					}
					writer.newLine();
				}
			}
			writer.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void storeTelevisionKVariables(float[] array, File outputFile) {
		BufferedWriter writer = null;
		try {
			writer = Utils.createBufferedWriter(outputFile);
			int index = 0;
			for (int groupId = 0; groupId < GROUP_NUM; groupId ++) {
				int K = groupMixtureNums[groupId];
				for (int innerTelevisionId = 0; innerTelevisionId <
						televisionNumInGroup[groupId]; innerTelevisionId ++) {
					for (int k = 0; k < K; k ++) {
						writer.write(String.valueOf(array[index]));
						index ++;
						if (k != K - 1) {
							writer.write(TVConstants.SPACER);
						}
					}
					if (index != TOTAL_LATENT_USER_NUM) {
						writer.newLine();
					}
				}
			}
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			Utils.cleanup(writer);
		}
	}

	private void normalize(float[] array, int startIndex, int len) throws TVException {
		float max        = 0;
		boolean[] status = new boolean[len];
		int invalidCount = 0;
		float sum = 0;
		for (int k = 0; k < len; k ++) {
			int index = startIndex + k;
			if (max < array[index]) max = array[index];
			if (Double.isInfinite(array[index]) ||
				Double.isNaN(array[index])) {
				invalidCount ++;
				status[k] = true;
			}
			sum += array[index];
		}
		if (max == 0) throw new TVException("normalize: max = " + max);
		
		if (invalidCount == 0) {
			sum = 0;
			for (int k = 0; k < len; k ++) {
				sum += (
					array[startIndex + k] /= max
				);
			}
		}
		if (Double.isInfinite(sum) || Double.isNaN(sum)) {
			if (invalidCount == 0) throw new TVException("InvalidCount = " + invalidCount);
			for (int k = 0; k < len; k ++) {
				if (status[k]) {
					array[startIndex + k] = 1.0f / invalidCount;
				} else {
					array[startIndex + k] = 0;
				}
			}
		} else {
			for (int k = 0; k < len; k ++) {
				array[startIndex + k] /= sum;
			}
		}
	}
}
