package com.uncc.tv.recommender;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.apache.commons.math3.linear.BlockRealMatrix;
import org.apache.commons.math3.linear.MatrixUtils;
import org.apache.commons.math3.linear.RealMatrix;

import com.lhy.tool.ToolException;
import com.lhy.tool.util.Utils;
import com.uncc.tv.TVConstants;
import com.uncc.tv.TVException;

/**
 * It shares some common methods among several classes.
 * 
 * @author Huayu Li
 *
 */
public class PMFBase {
	private static final String STRING_PROGRAM_NAME        = "ProgramName";
	private static final String STRING_LAMBDA              = "Lambda";
	private static final String STRING_LAMBDA_PROGRAM      = "LambdaProgram";
	private static final String STRING_LAMBDA_TELEVISION   = "LambdaTelevision";
	private static final String STRING_PROGRAM_NUM         = "ProgramNum";
	private static final String STRING_TELEVISION_NUM      = "TelevisionNum";
	private static final String STRING_FEATURE_NUM         = "FeatureNum";
	private static final String STRING_RATING_RECORD_NUM   = "RatingRecordNum";
	private static final String STRING_CONVERGE_THRESHOLD  = "ConvergeThreshold";
	protected static final String STRING_MEAN_TRAIN_RATING = "MeanTrainRating";

	protected float ERROR_THRESHOLD               = 1.0e-5f;
	protected boolean MINUS_MEAN_RATING           = true;
	protected static final boolean INIT_FROM_FILE = false;
	protected static final String FILE_NAME_PARAM              = "Param";
	protected static final String FILE_NAME_PROGRAM_FEATURE    = "Program_Feature";
	protected static final String FILE_NAME_TELEVISION_FEATURE = "Television_Feature";

	protected float[] programFeature    = null;
	protected float[] televisionFeature = null;
	protected float[] ratingsInProgramOrder    = null;
	protected float[] ratingsInTelevisionOrder = null;
	// the first element is 0, the element is the current group television start index
	protected int[] realUserIdMapToTelevisionIndex = null;
	protected int[] televisionNumInProgramOrder    = null;
	protected int[] televisionIndexInProgramOrder  = null;
	protected int[] programNumInTelevisionOrder    = null;
	protected int[] programIndexInTelevisionOrder  = null;

	protected int FEATURE_NUM;
	protected int PROGRAM_NUM;
	protected int TELEVISION_NUM;
	protected int RATING_RECORD_NUM;

	protected float lambdaProgram    = 0.01f;
	protected float lambdaTelevision = 0.01f;
	protected float lambda           = 0.01f;

	protected File outputPath = null;
	protected File initPath   = null;
	protected File trainFile  = null;
	protected File testFile   = null;
	protected File resultFile = null;

	protected double meanRating   = Double.NaN;
	protected boolean FORCE_CHECK = true;
	protected String PROGRAM_NAME = null;

	public PMFBase() {
		if (System.getProperty("NO_MINUS_MEAN_RATING") != null) {
			MINUS_MEAN_RATING = false;
			Utils.println("[Info] Won't minus mean rating.");
		} else {
			Utils.println("[Info] Will minus mean rating.");
		}
	}

	protected void updateProgramFeature() {
		int recordIndex  = 0;
		int programIndex = 0;
		for (int programId = 0; programId < PROGRAM_NUM; programId ++) {
			double[][] TT = new double[FEATURE_NUM][FEATURE_NUM];
			float[] V     = new float[FEATURE_NUM];
			for (int i = 0; i < televisionNumInProgramOrder[programId]; i++) {
				int televisionId = televisionIndexInProgramOrder[recordIndex];
				int televisionStartIndex = televisionId * FEATURE_NUM;
				for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
					float currentValue = televisionFeature[televisionStartIndex + featureId];
					V[featureId] += ratingsInProgramOrder[recordIndex] *
							currentValue;
					for (int f = 0; f < FEATURE_NUM; f ++) {
						TT[featureId][f] +=
							currentValue *
							televisionFeature[televisionStartIndex + f];
					}
				}
				recordIndex ++;
			}
			for (int i = 0; i < TT.length; i ++) {
				for (int j = 0; j < TT[0].length; j ++) {
					TT[i][j] = TT[i][j] * lambda;
				}
				TT[i][i] += lambdaProgram;
			}
			TT = inverse(TT);
			for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
				float value = 0;
				for (int i = 0; i < FEATURE_NUM; i ++) {
					value += TT[featureId][i] * V[i];
				}
				programFeature[programIndex] = value * lambda;
				programIndex ++;
			}
		}
	}

	/*
	 * The line format is as follows:
	 * 	UserId '\t' ProgramId '\t' RawRating
	 */
	protected double predict(File inputFile) throws TVException, IOException {
		if (! Utils.exists(inputFile)) {
			throw new TVException("File does not exist.[%s]",
				inputFile == null ? null: inputFile.getAbsolutePath());
		}
		BufferedReader reader = Utils.createBufferedReader(inputFile);
		String line           = null;
		double rmse           = 0;
		int recordNum         = 0;
		while ((line = reader.readLine()) != null) {
			if (Utils.isEmpty(line = line.trim())) continue;
			String array[] = line.split("\t");
			if (array.length != 3) {
				throw new TVException("Failed to parse line : " + line);
			}
			int realTelevisionId    = Integer.parseInt(array[0]);
			int programId           = Integer.parseInt(array[1]);
			float groundtruthRating = (float)(Float.parseFloat(array[2]));
			if (realTelevisionId < 0 || realTelevisionId >=
					realUserIdMapToTelevisionIndex.length) {
				throw new TVException("Cannot find the mapped id for the television %s.",
					realTelevisionId);
			}
			if (programId < 0 || programId >= PROGRAM_NUM) {
				throw new TVException("Cannot find the program %s.",
					programId);
			}
			int televisionId     = realUserIdMapToTelevisionIndex[realTelevisionId];
			double predictRating = 0;
			int televisionStartIndex = televisionId * FEATURE_NUM;
			int programStartIndex    = programId * FEATURE_NUM;
			for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
				predictRating +=
					televisionFeature[televisionStartIndex + featureId] *
					programFeature[programStartIndex + featureId];
			}
			if (MINUS_MEAN_RATING) {
				predictRating += meanRating;
			}
			rmse += (predictRating - groundtruthRating) *
					(predictRating - groundtruthRating);
			recordNum ++;
		}
		reader.close();

		return Math.sqrt(rmse / recordNum);
	}

	protected static double predict(final int TELEVISION_NUM, final int PROGRAM_NUM,
			final int FEATURE_NUM, float[] televisionFeature,
			float[] programFeature, int televisionId, int programId,
				float meanTrainRating) {
		return predict(TELEVISION_NUM, PROGRAM_NUM, FEATURE_NUM,
				televisionFeature, programFeature,
				televisionId, programId) + meanTrainRating;
	}

	protected static double predict(final int TELEVISION_NUM, final int PROGRAM_NUM,
			final int FEATURE_NUM, float[] televisionFeature,
			float[] programFeature, int televisionId, int programId) {
		int televisionStartIndex = televisionId * FEATURE_NUM;
		int programStartIndex    = programId * FEATURE_NUM;
		double predictRating     = 0;
		for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
			predictRating += televisionFeature[televisionStartIndex + featureId] *
					programFeature[programStartIndex + featureId];
		}
		return predictRating;
	}

	/*
	 * The line format is as follows:
	 * 	UserId '\t' ProgramId '\t' RawRating
	 * Before calling this method, must call parseTrainClusterFile()
	 */
	protected void parseTrainRatingFile(File trainRatingFile) throws TVException {
		BufferedReader reader = null;
		BufferedWriter writer = null;
		try {
			Utils.println("[Info] FORCE_CHECK = %s", FORCE_CHECK);

			// calculate mean rating
			meanRating = calculateMeanRating(trainRatingFile);
			Utils.println("[Info] MeanRating = " + meanRating);

			File tmpTrainRatingFile = new File(outputPath,
					"_tmp_" +  trainRatingFile.getName());
			if (MINUS_MEAN_RATING) {
				tmpTrainRatingFile.delete();
			}

			// load rating in program order
			Map<Integer, Map<Integer, Float>> programTelevisionRatingMap =
					new HashMap<Integer, Map<Integer, Float>>();
			String line = null;
			reader      = Utils.createBufferedReader(trainRatingFile);
			writer      = MINUS_MEAN_RATING ?
					Utils.createBufferedWriter(tmpTrainRatingFile):
					null;
			int minProgramId = Integer.MAX_VALUE;
			int maxProgramId = -1;
			RATING_RECORD_NUM = 0;
			while ((line = reader.readLine()) != null) {
				if (Utils.isEmpty(line = line.trim())) continue;
				String[] array = line.split("\t");
				if (array.length != 3) {
					reader.close();
					reader = null;
					throw new TVException("Failed to parse TrainRatingFile: %s", line);
				}
				int televisionId = Integer.parseInt(array[0]);
				int programId    = Integer.parseInt(array[1]);
				float rating     = Float.parseFloat(array[2]);
				if (MINUS_MEAN_RATING) {
					rating -= meanRating;
				}
				Map<Integer, Float> televisionRatingMap =
					programTelevisionRatingMap.get(programId);
				if (televisionRatingMap == null) {
					televisionRatingMap = new HashMap<Integer, Float>();
					televisionRatingMap.put(televisionId, rating);
					programTelevisionRatingMap.put(programId,
							televisionRatingMap);
				} else {
					if (televisionRatingMap.containsKey(televisionId)) {
						reader.close();
						reader = null;
						throw new TVException("Television %s has existed. Line : %s",
							televisionId, line);
					}
					televisionRatingMap.put(televisionId, rating);
				}
				if (programId < minProgramId) minProgramId = programId;
				if (programId > maxProgramId) maxProgramId = programId;
				RATING_RECORD_NUM ++;
				if (MINUS_MEAN_RATING) {
					Utils.write(writer, televisionId + "\t" +
							programId + "\t" +
							rating, true);
				}
			}
			reader.close();
			if (MINUS_MEAN_RATING) {
				writer.close();
			}
			if (programTelevisionRatingMap.size() != PROGRAM_NUM) {
				String errMsg = String.format("Failed to parse TrainRatingFile: program number does not match with predefined number." +
						"[PredefinedProgramNum=%s][ParsedProgramNum=%s]",
						PROGRAM_NUM, programTelevisionRatingMap.size());
				if (FORCE_CHECK) {
					throw new TVException(errMsg);
				} else {
					Utils.err(errMsg);
				}
			}
			if (minProgramId != 0) {
				String errMsg = String.format("MinProgramId=%s, instead of 0.",
						minProgramId);
				if (FORCE_CHECK) {
					throw new TVException(errMsg);
				} else {
					Utils.err(errMsg);
				}
			}
			if (maxProgramId != PROGRAM_NUM - 1) {
				String errMsg = String.format("MaxProgramId=%s, instead of %s.",
						maxProgramId, PROGRAM_NUM - 1);
				if (FORCE_CHECK) {
					throw new TVException(errMsg);
				} else {
					Utils.err(errMsg);
				}
			}
			{
				ratingsInProgramOrder         = new float[RATING_RECORD_NUM];
				televisionIndexInProgramOrder = new int[RATING_RECORD_NUM];
				televisionNumInProgramOrder   = new int[PROGRAM_NUM];
				int[] programIds = getSortedKeys(programTelevisionRatingMap.keySet());
				int index        = 0;
				for (int programId : programIds) {
					Map<Integer, Float> televisionRatingMap =
						programTelevisionRatingMap.get(programId);
					int[] televisionIds = getSortedKeys(
						televisionRatingMap.keySet());
					for (int televisionId : televisionIds) {
						ratingsInProgramOrder[index] =
							televisionRatingMap.get(televisionId);
						televisionIndexInProgramOrder[index] =
							realUserIdMapToTelevisionIndex[televisionId];
						index ++;
					}
					televisionNumInProgramOrder[programId] =
						televisionRatingMap.size();
				}
			}
			// clear
			programTelevisionRatingMap = null;

			// load rating in television order
			if (MINUS_MEAN_RATING) {
				reader = Utils.createBufferedReader(tmpTrainRatingFile);
			} else {
				reader = Utils.createBufferedReader(trainRatingFile);
			}
			Map<Integer, Map<Integer, Float>> televisionProgramRatingMap =
				new HashMap<Integer, Map<Integer, Float>>();
			int minTelevisionId = Integer.MAX_VALUE;
			int maxTelevisionId = -1;
			while ((line = reader.readLine()) != null) {
				if (Utils.isEmpty(line = line.trim())) continue;
				String[] array = line.split("\t");
				if (array.length != 3) {
					reader.close();
					reader = null;
					throw new TVException("Failed to parse TrainRatingFile: %s", line);
				}
				int televisionId = realUserIdMapToTelevisionIndex
						[Integer.parseInt(array[0])];
				int programId    = Integer.parseInt(array[1]);
				float rating     = Float.parseFloat(array[2]);
				Map<Integer, Float> programRatingMap =
					televisionProgramRatingMap.get(televisionId);
				if (programRatingMap == null) {
					programRatingMap = new HashMap<Integer, Float>();
					programRatingMap.put(programId, rating);
					televisionProgramRatingMap.put(televisionId,
						programRatingMap);
				} else {
					if (programRatingMap.containsKey(programId)) {
						throw new TVException("Program %s has existed. Line:%s",
							programId, line);
					}
					programRatingMap.put(programId, rating);
				}
				if (televisionId < minTelevisionId)
					minTelevisionId = televisionId;
				if (televisionId > maxTelevisionId)
					maxTelevisionId = televisionId;
			} // end
			reader.close();
			if (MINUS_MEAN_RATING) {
				tmpTrainRatingFile.delete();
			}
			if (televisionProgramRatingMap.size() != TELEVISION_NUM) {
				String errMsg = String.format("Failed to parse TrainRatingFile : television number does not match with the predefined number." +
						"[PredefinedTelevisionNum=%s][LoadedTelevisionNum=%s]",
						TELEVISION_NUM, televisionProgramRatingMap.size());
				if (FORCE_CHECK) {
					throw new TVException(errMsg);
				} else {
					Utils.err(errMsg);
				}
			}
			if (minTelevisionId != 0) {
				String errMsg = String.format("MinTelevisionNum=%s, instead of 0.",
						minTelevisionId);
				if (FORCE_CHECK) {
					throw new TVException(errMsg);
				} else {
					Utils.err(errMsg);
				}
			}
			if (maxTelevisionId != TELEVISION_NUM - 1) {
				String errMsg = String.format("MaxTelevisionNum=%s, instead of %s.",
						maxTelevisionId, TELEVISION_NUM - 1);
				if (FORCE_CHECK) {
					throw new TVException(errMsg);
				} else {
					Utils.err(errMsg);
				}
			}
			{
				ratingsInTelevisionOrder      = new float[RATING_RECORD_NUM];
				programIndexInTelevisionOrder = new int[RATING_RECORD_NUM]; 
				programNumInTelevisionOrder   = new int[TELEVISION_NUM];
				int[] televisionIds = getSortedKeys(
						televisionProgramRatingMap.keySet());
				int index = 0;
				for (int televisionId : televisionIds) {
					Map<Integer, Float> programRatingMap =
						televisionProgramRatingMap.get(televisionId);
					int[] programIds = getSortedKeys(
						programRatingMap.keySet());
					for (int programId : programIds) {
						ratingsInTelevisionOrder[index] =
							programRatingMap.get(programId);
						programIndexInTelevisionOrder[index] =
							programId;
						index ++;
					}
					programNumInTelevisionOrder[televisionId] =
						programRatingMap.size();
				}
			}
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			Utils.cleanup(reader);
			Utils.cleanup(writer);
		}
	}

	protected Properties getParams() {
		Properties props = new Properties();
		props.put(STRING_LAMBDA, 		String.valueOf(lambda));
		props.put(STRING_LAMBDA_PROGRAM, 	String.valueOf(lambdaProgram));
		props.put(STRING_LAMBDA_TELEVISION,	String.valueOf(lambdaTelevision));
		props.put(STRING_PROGRAM_NUM,		String.valueOf(PROGRAM_NUM));
		props.put(STRING_TELEVISION_NUM,	String.valueOf(TELEVISION_NUM));
		props.put(STRING_FEATURE_NUM,		String.valueOf(FEATURE_NUM));
		props.put(STRING_RATING_RECORD_NUM,	String.valueOf(RATING_RECORD_NUM));
		props.put(STRING_MEAN_TRAIN_RATING,	String.valueOf(meanRating));
		props.put(STRING_PROGRAM_NAME,		PROGRAM_NAME);
		props.put(STRING_CONVERGE_THRESHOLD, 	String.valueOf(ERROR_THRESHOLD));
		props.put("MINUS_MEAN_RATING",		String.valueOf(MINUS_MEAN_RATING));
		return props;
	}

	public static double[][] inverse(double[][] matrix) {
		return inverse(new BlockRealMatrix(matrix)).getData();
	}

	public static RealMatrix inverse(RealMatrix matrix) {
		int splitIndex = matrix.getColumnDimension() > matrix.getRowDimension() ?
					matrix.getRowDimension() : matrix.getColumnDimension();
		splitIndex = splitIndex == 2 ? 0 : splitIndex / 2;
		return MatrixUtils.blockInverse(matrix, splitIndex);
	}

	private double calculateMeanRating(File inputFile) throws TVException{
		if (! Utils.exists(inputFile)) {
			throw new TVException("Input file does not exist.[%s]",
				inputFile == null ? null : inputFile.getAbsolutePath());
		}
		BufferedReader reader = null;
		int recordNum         = 0;
		double meanRating     = 0;
		try {
			reader      = Utils.createBufferedReader(inputFile);
			String line = null;
			while ((line = reader.readLine()) != null) {
				if (Utils.isEmpty(line = line.trim())) continue;
				String[] array = line.split("\t");
				if (array.length != 3) {
					reader.close();
					reader = null;
					throw new TVException("Failed to parse TrainRatingFile: %s", line);
				}
				meanRating += Float.parseFloat(array[2]);
				recordNum ++;
			}
			meanRating = recordNum == 0 ? 0 : (meanRating / (recordNum + 0.0));
			return meanRating;
		} catch (IOException e) {
			e.printStackTrace();
			throw new TVException(e.toString());
		} finally{
			Utils.cleanup(reader);
		}
	}

	protected void initProgramFeature(File programFeatureFile) throws TVException {
		if (programFeatureFile == null) {
			programFeature = new float[PROGRAM_NUM * FEATURE_NUM];
			for (int i = 0; i < programFeature.length; i ++) {
				programFeature[i] = (float)(Math.random() * 0.1f);
			}
		} else {
			try {
				//File programFeatureFile = new File(initPath, FILE_NAME_PROGRAM_FEATURE);
				Utils.println("[Info] Loading programFeature from " + programFeatureFile.getAbsolutePath());
				programFeature = convert2DTo1D(Utils.load2FloatArray(
					programFeatureFile, TVConstants.SPACER));
				if (programFeature.length != PROGRAM_NUM * FEATURE_NUM) {
					throw new TVException("Failed to load programFeature: dimension = %s which should be %s.",
						programFeature.length, PROGRAM_NUM * FEATURE_NUM);
				}
			} catch (ToolException e) {
				throw new TVException(e.toString());
			}
		}
	}

	protected void initTelevisionFeature(File televisionFeatureFile)
					throws TVException {
		if (televisionFeatureFile == null) {
			televisionFeature = new float[TELEVISION_NUM * FEATURE_NUM];
			for (int i = 0; i < televisionFeature.length; i ++) {
				televisionFeature[i] = (float)(Math.random() * 0.1f);
			}
		} else {
			try {
				//File televisionFeatureFile = new File(initPath, FILE_NAME_TELEVISION_FEATURE);
				Utils.println("[Info] Loading televisionFeature from " + televisionFeatureFile.getAbsolutePath());
				televisionFeature = convert2DTo1D(Utils.load2FloatArray(
					televisionFeatureFile, TVConstants.SPACER));
				if (televisionFeature.length != TELEVISION_NUM * FEATURE_NUM) {
					throw new TVException("Failed to load teleivsionFeature: dimension = %s which should be %s.",
						televisionFeature.length, TELEVISION_NUM * FEATURE_NUM);
				}
			} catch (ToolException e) {
				throw new TVException(e.toString());
			}
		}
	}

	protected void storeParams(String comment) {
		Properties params = getParams();
		params.list(System.out);
		try {
			if (!outputPath.exists()) {
				outputPath.mkdirs();
			}
			params.store(new FileWriter(new File(outputPath,
					FILE_NAME_PARAM)), comment);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/*
	 * ********************************************************************
	 * The following is the basic methods.
	 * ********************************************************************
	 */

	public static float[] convert2DTo1D(float[][] matrix) {
		if (matrix == null) return null;
		float[] array = new float[matrix.length * matrix[0].length];
		int index = 0;
		for (int row = 0; row < matrix.length; row ++) {
			for (int column = 0; column < matrix[row].length; column ++) {
				array[index] = matrix[row][column];
				index ++;
			}
		}
		return array;
	}
	
	private static int[] getSortedKeys(Set<Integer> keys) {
		int[] array = new int[keys.size()];
		int index   = 0;
		for (Integer key : keys) {
			array[index] = key;
			index ++;
		}
		Arrays.sort(array);
		return array;
	}

	protected void outputOneLine(BufferedWriter writer, int[] array)
							throws IOException {
		for (int i = 0; i < array.length; i ++) {
			writer.write(i + ":" + array[i]);
			if (i != array.length - 1)
				writer.write(TVConstants.SPACER);
		}
		writer.newLine();
	}

	protected void outputOneLine(BufferedWriter writer, float[] array)
							throws IOException {
		for (int i = 0; i < array.length; i ++) {
			writer.write(i + ":" + array[i]);
			if (i != array.length - 1)
				writer.write(TVConstants.SPACER);
		}
		writer.newLine();
	}

	protected void store2DArray(float[] array, int columnNum, File outputFile) {
		BufferedWriter writer = null;
		try {
			writer    = Utils.createBufferedWriter(outputFile);
			int count = 0;
			for (int i = 0; i < array.length; i ++) {
				writer.write(String.valueOf(array[i]));
				count ++;
				if (i != array.length - 1) {
					if (count == columnNum) {
						count = 0;
						writer.newLine();
					} else {
						writer.write(TVConstants.SPACER);
					}
				}
			}
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			Utils.cleanup(writer);
		}
	}
}
