package com.uncc.tv.recommender;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;

import com.lhy.tool.util.Utils;
import com.uncc.tv.ParamManager;
import com.uncc.tv.TVConstants;
import com.uncc.tv.TVException;

/**
 * This is weighted Mixture Probabilistic Matrix Factorization.
 * @author huayuli
 *
 */
public class WGPMF extends GPMF {
	private float epsilon     = 1.0e8f;
	private float ratingTimes = 40;
	private float[] convertedRatingsInProgramOrder    = null;
	private float[] convertedRatingsInTelevisionOrder = null;
	private float[] timesRatingsInProgramOrder        = null;
	private float[] timesRatingsInTelevisionOrder     = null;
	private float[] televisionTelevisionSum = null;
	private float[] programProgramSum       = null;

	public WGPMF(ParamManager paramManager) throws TVException, IOException{
		super(paramManager, "WGPMF", true);
		init();
	}

	@Override
	public void estimate() throws TVException, IOException {
		BufferedWriter resultWriter = Utils.createBufferedWriter(resultFile, true);
		getParams().store(resultWriter, String.format("%s Parameters", PROGRAM_NAME));

		double oldlowerbound  = calculateLowerbound();
		double errorThreshold = Double.MAX_VALUE;
		int iteration         = 1;
		if (testFile != null) {
			Utils.writeAndPrint(resultWriter, String.format(
				"Train RMSE = %s, Test RMSE = %s",
				predict(trainFile), predict(testFile)), true);
		} else {
			Utils.writeAndPrint(resultWriter,
				"Train RMSE = " + predict(trainFile), true);
		}
		do {
			updateProgramFeature();
			updateTelevisionFeature();
			updateWeight();
			updatePhi();
			updateMu();
			//if (iteration == 1) {
				//updateSigma();
			//}
			double lowerbound = calculateLowerbound();
			errorThreshold    = Math.abs(Math.abs((oldlowerbound - lowerbound) / oldlowerbound));
			oldlowerbound     = lowerbound;
			if (testFile != null) {
				Utils.writeAndPrint(resultWriter, String.format(
					"[Iteration = %s] LowerBound = %s, ErrorThreshold = %s, TrainRMSE = %s, TestRMSE = %s",
					iteration, lowerbound, errorThreshold,
					predict(trainFile), predict(testFile)),
					true);
			} else {
				Utils.writeAndPrint(resultWriter, String.format(
					"[Iteration = %s] LowerBound = %s, ErrorThreshold = %s, TrainRMSE = %s",
					iteration, lowerbound, errorThreshold,
					predict(trainFile)),
					true);
			}
			storeLatentVariables(outputPath, true);
			if (iteration >= 100) break;
			iteration ++;
		} while (iteration < 5 || errorThreshold > ERROR_THRESHOLD);
		storeLatentVariables(outputPath, true);
		resultWriter.close();
	}

	@Override
	protected void updateTelevisionFeature() throws IOException, TVException {
		BufferedReader sigmaReader = Utils.createBufferedReader(sigmaFile);
		int televisionIndex = 0;
		int televisionId    = 0;
		int recordIndex     = 0;
		int phiIndex        = 0;
		int muStartIndex    = 0;
		float[][] sigma     = null;
		reset(televisionTelevisionSum);
		for (int groupId = 0; groupId < GROUP_NUM; groupId ++) {
			int K = groupMixtureNums[groupId];
			sigma = new float[K][FEATURE_NUM * FEATURE_NUM];
			for (int k = 0; k < K; k ++) {
				sigma[k] = readDataInLine(sigmaReader);
			}
			for (int innerTelevisionId = 0; innerTelevisionId <
					televisionNumInGroup[groupId]; innerTelevisionId ++) {
				double[][] VVT = new double[FEATURE_NUM][FEATURE_NUM];
				float[]      T = new float[FEATURE_NUM];
				for (int i = 0; i < programNumInTelevisionOrder[televisionId]; i ++) {
					int programStartIndex =
						programIndexInTelevisionOrder[recordIndex] *
						FEATURE_NUM;
					for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
						float currentValue = programFeature[programStartIndex + featureId];
						T[featureId] += 2 * convertedRatingsInTelevisionOrder[recordIndex] *
								currentValue;
						for (int f = 0; f < FEATURE_NUM; f ++) {
							VVT[featureId][f] += currentValue *
								programFeature[programStartIndex + f] *
								timesRatingsInTelevisionOrder[recordIndex];
						}
					}
					recordIndex ++;
				}
				int index = 0;
				for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
					for (int f = 0; f < FEATURE_NUM; f ++) {
						VVT[featureId][f] = VVT[featureId][f] * 2 + programProgramSum[index];
						index ++;
					}
					VVT[featureId][featureId] += lambdaRegT;
				}
				for (int k = 0; k < K; k ++) {
					int tmpMuStartIndex = muStartIndex + k * FEATURE_NUM;
					for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
						float sum           = 0;
						int sigmaStartIndex = featureId * FEATURE_NUM;
						for (int f = 0; f < FEATURE_NUM; f ++) {
							float value = varPhi[phiIndex] * sigma[k][sigmaStartIndex + f];
							VVT[featureId][f] += value;
							sum += value * varMu[tmpMuStartIndex + f];
						}
						T[featureId] += sum;
					}
					phiIndex ++;
				}
				VVT = inverse(VVT);
				for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
					float v = 0;
					for (int f = 0; f < FEATURE_NUM; f ++) {
						v+= VVT[featureId][f] * T[f];
					}
					televisionFeature[televisionIndex + featureId] = v;
				}
				index = 0;
				for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
					float v = televisionFeature[televisionIndex + featureId];
					for (int f = 0; f < FEATURE_NUM; f ++) {
						televisionTelevisionSum[index] += v * televisionFeature[televisionIndex + f];
						index ++;
					}
				}
				televisionIndex += FEATURE_NUM;
				televisionId ++;
			} // end for (int innerTelevisionId = 0
			muStartIndex += K * FEATURE_NUM;
		} // end for (int groupId = 0;
		for (int i = televisionTelevisionSum.length - 1; i >= 0; i --)
			televisionTelevisionSum[i] *= 2;
		String line = null;
		if ((line = sigmaReader.readLine()) != null)
			throw new TVException("Sigma file does not end. NextLine:%s", line);
		sigmaReader.close();
	}

	@Override
	protected void updateProgramFeature() {
		int recordIndex  = 0;
		int programIndex = 0;
		reset(programProgramSum);
		for (int programId = 0; programId < PROGRAM_NUM; programId ++) {
			double[][] TT = new double[FEATURE_NUM][FEATURE_NUM];
			float[] V     = new float[FEATURE_NUM];
			for (int i = 0; i < televisionNumInProgramOrder[programId]; i++) {
				int televisionId = televisionIndexInProgramOrder[recordIndex];
				int televisionStartIndex = televisionId * FEATURE_NUM;
				for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
					float currentValue = televisionFeature[televisionStartIndex + featureId];
					V[featureId] += convertedRatingsInProgramOrder[recordIndex] *
							currentValue;
					for (int f = 0; f < FEATURE_NUM; f ++) {
						TT[featureId][f] +=
							currentValue *
							televisionFeature[televisionStartIndex + f] *
							timesRatingsInProgramOrder[recordIndex];
					}
				}
				recordIndex ++;
			}
			int index = 0;
			for (int i = 0; i < TT.length; i ++) {
				for (int j = 0; j < TT[0].length; j ++) {
					TT[i][j] = TT[i][j] * 2 + televisionTelevisionSum[index];
					index ++;
				}
				TT[i][i] += lambdaProgram;
				V[i] *= 2;
			}
			TT = inverse(TT);
			for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
				float value = 0;
				for (int i = 0; i < FEATURE_NUM; i ++) {
					value += TT[featureId][i] * V[i];
				}
				programFeature[programIndex + featureId] = value;
			}
			index = 0;
			for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
				float v = programFeature[programIndex + featureId];
				for (int f = 0; f < FEATURE_NUM; f ++) {
					programProgramSum[index] += v * programFeature[programIndex + f];
					index ++;
				}
			}
			programIndex += FEATURE_NUM;
		}
		for (int i = programProgramSum.length - 1; i >= 0; i --)
			programProgramSum[i] *= 2;
	}

	protected double predict(File inputFile) throws TVException, IOException {
		if (! Utils.exists(inputFile)) {
			throw new TVException("File does not exist.[%s]",
				inputFile == null ? null: inputFile.getAbsolutePath());
		}
		BufferedReader reader = Utils.createBufferedReader(inputFile);
		String line           = null;
		double rmse           = 0;
		int recordNum         = 0;
		while ((line = reader.readLine()) != null) {
			if (Utils.isEmpty(line = line.trim())) continue;
			String array[] = line.split("\t");
			if (array.length != 3) {
				throw new TVException("Failed to parse line : " + line);
			}
			int realTelevisionId = Integer.parseInt(array[0]);
			int programId        = Integer.parseInt(array[1]);
			if (realTelevisionId < 0 || realTelevisionId >=
					realUserIdMapToTelevisionIndex.length) {
				throw new TVException("Cannot find the mapped id for the television %s.",
					realTelevisionId);
			}
			if (programId < 0 || programId >= PROGRAM_NUM) {
				throw new TVException("Cannot find the program %s.",
					programId);
			}
			int televisionId     = realUserIdMapToTelevisionIndex[realTelevisionId];
			double predictRating = 0;
			int televisionStartIndex = televisionId * FEATURE_NUM;
			int programStartIndex    = programId * FEATURE_NUM;
			for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
				predictRating +=
					televisionFeature[televisionStartIndex + featureId] *
					programFeature[programStartIndex + featureId];
			}
			rmse += (predictRating - 1) * (predictRating - 1);
			recordNum ++;
		}
		reader.close();

		return Math.sqrt(rmse / recordNum);
	}

	@Override
	protected double calculateLowerbound() throws IOException, TVException {
		double lbvalue   = 0;
		// (R - T'V)_2
		int recordIndex = 0;
		for (int televisionId = 0; televisionId < TELEVISION_NUM; televisionId ++) {
			int televisionStartIndex = televisionId * FEATURE_NUM;
			boolean status[]         = new boolean[PROGRAM_NUM];
			for (int i = 0; i < programNumInTelevisionOrder[televisionId]; i ++) {
				double v      = 0;
				int programId = programIndexInTelevisionOrder[recordIndex];
				int programStartIndex = programId * FEATURE_NUM;
				status[programId]     = true;
				for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {					
					v += televisionFeature[televisionStartIndex + featureId] *
						programFeature[programStartIndex + featureId];
				}
				lbvalue -= convertedRatingsInTelevisionOrder[recordIndex] *
						(1 - v) * (1 - v);
				recordIndex ++;
			}
			for (int programId = 0; programId < PROGRAM_NUM; programId ++) {
				if (status[programId]) continue;
				double v              = 0;
				int programStartIndex = programId * FEATURE_NUM;
				for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {					
					v += televisionFeature[televisionStartIndex + featureId] *
						programFeature[programStartIndex + featureId];
				}
				lbvalue -= v * v;
			}
		}

		// V'V
		double sum       = 0;
		int programIndex = 0;
		for (int programId = 0; programId < PROGRAM_NUM; programId ++) {
			for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
				sum += programFeature[programIndex] *
					programFeature[programIndex];
				programIndex ++;
			}
		}
		lbvalue += -0.5 * lambdaProgram * sum;
		if (Double.isNaN(lbvalue) || Double.isInfinite(lbvalue)) {
			Utils.err("V'V : lbvalue = " + lbvalue);
		}

		// T
		BufferedReader sigmaReader = Utils.createBufferedReader(sigmaFile);
		double phiSum               = 0;
		double phiLnPhiSum          = 0;
		double phiLnWeightSum       = 0;
		double phiLnSigmaDetSqrtSum = 0;
		double lnSigmaDetSqrtSum    = 0;
		double lnWeightSum          = 0;
		double muTmu                = 0;
		double trTW0Sigma           = 0;
		double T_MuTT_Mu            = 0;
		int phiStartIndex           = 0;
		int lkIndex                 = 0;
		int muStartIndex            = 0;
		int televisionStartIndex    = 0;
		for (int groupId = 0; groupId < GROUP_NUM; groupId ++) {
			int K = groupMixtureNums[groupId];
			for (int k = 0; k < K; k ++) {
				float[] sigma = readDataInLine(sigmaReader);
				for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
					//muTmu      += varMu[muStartIndex + featureId] * varMu[muStartIndex + featureId];
					for (int f = 0; f < FEATURE_NUM; f ++) {
						muTmu += varMu[muStartIndex + f] *
							sigma[f * FEATURE_NUM + featureId] *
							varMu[muStartIndex + featureId];
					}
					trTW0Sigma += W0_Inverse[featureId] *
						sigma[featureId * FEATURE_NUM + featureId];
				}
				double phiSumTelevision = 0;
				for (int innerTelevisionId = 0; innerTelevisionId <
						televisionNumInGroup[groupId]; innerTelevisionId ++) {
					int phiIndex = phiStartIndex + innerTelevisionId * K + k;
					double phi        = varPhi[phiIndex];
					double weight     = varWeight[phiIndex];
					if (weightAssignments == null ||
							weightAssignments[phiIndex]) {
						phiLnWeightSum   += phi * Math.log(weight);
						lnWeightSum      += Math.log(weight);
						phiSumTelevision += phi;
						phiLnPhiSum      += phi * Math.log(phi);
					}
					//
					float[] diff = new float[FEATURE_NUM];
					int tIndex   = televisionStartIndex + innerTelevisionId * FEATURE_NUM;
					for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
						diff[featureId] = televisionFeature[tIndex + featureId] -
								varMu[muStartIndex + featureId];
					}
					double TTsum = 0;
					for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
						double v = 0;
						for (int f = 0; f < FEATURE_NUM; f ++) {
							v += diff[f] * sigma[f * FEATURE_NUM + featureId];
						}
						TTsum += v * diff[featureId] ;
					}
					T_MuTT_Mu += TTsum * phi;
				}
				lnSigmaDetSqrtSum    += sigmaDetSqrt[lkIndex];
				phiSum               += phiSumTelevision;
				phiLnSigmaDetSqrtSum += sigmaDetSqrt[lkIndex] *
						phiSumTelevision;
				muStartIndex         += FEATURE_NUM;
				lkIndex ++;
			} // end for k
			phiStartIndex        += televisionNumInGroup[groupId] * K;
			televisionStartIndex += televisionNumInGroup[groupId] * FEATURE_NUM;
		}
		lbvalue += -0.5 * FEATURE_NUM * Math.log(2 * Math.PI) * phiSum;
		if (Double.isNaN(lbvalue) || Double.isInfinite(lbvalue)) {
			Utils.err("-0.5 * FEATURE_NUM * Math.log(2 * Math.PI) * phiSum : lbvalue = " + lbvalue);
		}
		lbvalue += (alpha - 1) * lnWeightSum;
		if (Double.isNaN(lbvalue) || Double.isInfinite(lbvalue)) {
			Utils.err("(alpha - 1) * Math.log(lnWeightSum) : lbvalue=%s, lnWeightSum=%s", lbvalue, lnWeightSum);
		}
		lbvalue += phiLnWeightSum;
		lbvalue -= phiLnPhiSum;
		lbvalue += (v0_D_1 + 1) * lnSigmaDetSqrtSum;
		if (Double.isNaN(lbvalue) || Double.isInfinite(lbvalue)) {
			Utils.err("v0_D_1 * lnSigmaDetSqrtSum : lbvalue = " + lbvalue);
		}
		lbvalue -= 0.5 * muTmu * lambdaTelevision;
		if (lbvalue > 0) Utils.err("v0_D_1 * lnSigmaDetSqrtSum = " + lbvalue);
		lbvalue += phiLnSigmaDetSqrtSum;
		if (Double.isNaN(lbvalue) || Double.isInfinite(lbvalue)) {
			Utils.err("0.5 * v0_D_1 * Math.log(lnSigmaDetSqrtSum) : lbvalue = " + lbvalue);
		}
		lbvalue -= 0.5 * trTW0Sigma;
		lbvalue -= 0.5 * T_MuTT_Mu;
		String line = null;
		if ((line = sigmaReader.readLine()) != null) 
			throw new TVException("Sigma file does not end. NextLine:%s", line);
		sigmaReader.close();
		return lbvalue;
	}

	private void init() {
		convertedRatingsInProgramOrder = new float[ratingsInProgramOrder.length];
		timesRatingsInProgramOrder     = new float[ratingsInProgramOrder.length];
		for (int i = convertedRatingsInProgramOrder.length - 1; i >= 0; i --) {
			float v = (float)(ratingTimes * Math.log(1 + ratingsInProgramOrder[i] * epsilon));
			convertedRatingsInProgramOrder[i] = v + 1;
			timesRatingsInProgramOrder[i]     = v;
		}
		convertedRatingsInTelevisionOrder = new float[ratingsInTelevisionOrder.length];
		timesRatingsInTelevisionOrder     = new float[ratingsInTelevisionOrder.length];
		for (int i = convertedRatingsInTelevisionOrder.length - 1; i >= 0; i --) {
			float v = (float)(ratingTimes * Math.log(1 + ratingsInProgramOrder[i] * epsilon));
			convertedRatingsInTelevisionOrder[i] = v + 1;
			timesRatingsInTelevisionOrder[i]     = v;
		}
		// init sum_T'T
		televisionTelevisionSum  = new float[FEATURE_NUM * FEATURE_NUM];
		int televisionStartIndex = 0;
		for (int televisionId = 0; televisionId < TELEVISION_NUM; televisionId ++) {
			int index = 0;
			for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
				float v = televisionFeature[televisionStartIndex + featureId];
				for (int f = 0; f < FEATURE_NUM; f ++) {
					televisionTelevisionSum[index] += v * televisionFeature[televisionStartIndex + f];
					index ++;
				}
			}
			televisionStartIndex += FEATURE_NUM;
		}
		for (int i = televisionTelevisionSum.length - 1; i >= 0; i --) {
			televisionTelevisionSum[i] *= 2;
		}
		// init sum_V'V
		programProgramSum     = new float[FEATURE_NUM * FEATURE_NUM];
		int programStartIndex = 0;
		for (int programId = 0; programId < PROGRAM_NUM; programId ++) {
			int index = 0;
			for (int featureId = 0; featureId < FEATURE_NUM; featureId ++) {
				float v = programFeature[programStartIndex + featureId];
				for (int f = 0; f < FEATURE_NUM; f ++) {
					programProgramSum[index] += v * programFeature[programStartIndex + f];
					index ++;
				}
			}
			programStartIndex += FEATURE_NUM;
		}
		for (int i = programProgramSum.length - 1; i >= 0; i --) {
			programProgramSum[i] *= 2;
		}
	}

	private void reset(float[] array) {
		for (int i = array.length - 1; i >= 0; i --)
			array[i] = 0;
	}
}
